--[[
@description raw metering intervalometer
@chdk_version 1.4.0.3862
#ui_shots=5 "Shots"
#ui_interval_s10=20 "Interval Sec/10"
#ui_meter_size=1200 "Meter size" [8 5000]
#ui_meter_step=5 "Meter step"
#ui_ev_change_f=3 "Max ev change 1/x" [1 16]
#ui_tv_max_s100=100 "Max Tv Sec/100"
#ui_sv_target_mkt=80 "Target ISO"
#ui_tv_sv_adj_s100=25 "ISO adj Tv Sec/100"
#ui_sv_max_mkt=800 "Max ISO"
#ui_exp_over_thresh_pct=3 "Overexp thresh %"
#ui_exp_over_margin_ev_f=16 "Overexp ev 1/x" [1 96]
#ui_exp_under_thresh_pct=10 "Underexp thresh %"
#ui_exp_under_margin_ev=5 "Underexp ev < neutral" [2 8]
#ui_display_mode_t=0 "Display" {On Off Blt_Off} table
#ui_shutdown_finish=false "Shutdown on finish"
#ui_shutdown_lowbat=true "Shutdown on low battery"
#ui_shutdown_lowspace=true "Shutdown on low space"
#ui_do_draw=true "Draw debug info"
#ui_log_append=false "Append log"
#ui_sim=false "Run simulation"

Version 0.5

Copyright reyalp (at) gmail.com
License: GPL
]]
require'hookutil'
require'rawoplib'
props=require'propcase'

interval=ui_interval_s10*100

-- log module
log={}
function log:init(opts)
	if not opts then
		error('missing opts');
	end
	self.cols={unpack(opts.cols)}
	self.vals={}
	self.funcs={}
	self.tables={}
	if opts.funcs then
		for n,f in pairs(opts.funcs) do
			if type(f) ~= 'function' then
				error('expected function')
			end
			self.funcs[n] = f
		end
	end
	-- TODO may accept other options than sep later
	if opts.tables then
		for n,sep in pairs(opts.tables) do
			self.tables[n] = {sep=sep}
		end
	end
	self:reset_vals()
	-- checks after vals initialized
	for n, v in pairs(self.funcs) do
		if not self.vals[n] then
			error('missing func col '.. tostring(n))
		end
	end
	for n, v in pairs(self.tables) do
		if not self.vals[n] then
			error('missing table col '.. tostring(n))
		end
	end
	-- TODO name should accept autonumber or date based options
	if opts.append then
		self.fh = io.open(opts.name,'ab')
	else
		self.fh = io.open(opts.name,'wb')
	end
	if not self.fh then
		error('failed to open log')
	end
	self:write_csv(self.cols)
end
function log:write_csv(data)
	-- TODO should handle CSV quoting
	self.fh:write(string.format("%s\n",table.concat(data,',')))
end
function log:write()
	local data={}
	for i,name in ipairs(self.cols) do
		local v
		if self.funcs[name] then
			v=tostring(self.funcs[name]())
		elseif self.tables[name] then
			v=table.concat(self.vals[name],self.tables[name].sep)
		else
			v=self.vals[name]
		end
		table.insert(data,v)
	end
	self:write_csv(data)
	self:reset_vals()
end
function log:reset_vals()
	for i,name in ipairs(self.cols) do
		if self.tables[name] then
			self.vals[name] = {}
		else
			self.vals[name] = ''
		end
	end
end
function log:set(vals)
	for name,v in pairs(vals) do
		if not self.vals[name] then
			error("unknown log col "..tostring(name))
		end
		if self.funcs[name] then
			error("tried to set func col "..tostring(name))
		end
		if self.tables[name] then
			table.insert(self.vals[name],v)
		else
			self.vals[name] = tostring(v)
		end
	end
end
--[[
return a function that records time offset from col named base_name
if name is not provided, function expects target aname as arg
]]
function log:dt_logger(base_name,name)
	if not self.vals[base_name] then
		error('invalid base field name')
	end
	if not name then
		return function(name)
			if not self.vals[name] then
				error('invalid col name')
			end
			self.vals[name]=get_tick_count() - self.vals[base_name]
		end
	end
	if not self.vals[name] then
		error('invalid col name')
	end
	return function()
		self.vals[name]=get_tick_count() - self.vals[base_name]
	end
end

--[[
return a printf-like function that appends to table col
]]
function log:text_logger(name)
	if not self.vals[name] then
		error('invalid col name')
	end
	if not self.tables[name] then
		error('text logger must be table field '..tostring(name))
	end
	return function(fmt,...)
		table.insert(self.vals[name],string.format(fmt,...))
	end
end

function log:close()
	if self.fh then
		self.fh:close()
	end
end
-- end log module

-- display control module
disp={
	state=true,
	shutoff_time=0,
	mode=string.lower(ui_display_mode_t[ui_display_mode_t.value]),
}

function disp:init()
	if self.mode == 'on' then
		self._control_fn=function() end
	elseif self.mode == 'blt_off' then
		self.control_fn=set_backlight
	else
		self.control_fn=set_lcd_display
	end
end

function disp:update()
	-- toggled on, not expired, do nothing (should have been turned on in toggle)
	if self.shutoff_time > get_tick_count() then
		return
	end
	
	-- if on, and timeout expired turn off
	if self.state then
		self:enable(false)
		return
	end
	-- turn off every shot in backlight mode
	if self.mode == 'blt_off' then
		self:enable(false)
	end
end

function disp:enable(state)
	if self.mode == 'on' then
		return
	end
	self.state=state
	self.control_fn(state)
end

function disp:toggle(timeout)
	if not timeout then
		timeout = 30000
	end
	if self.state then
		self.shutoff_time = 0
		self:enable(false)
	else
		self.shutoff_time = get_tick_count() + timeout
		self:enable(true)
	end
end
disp:init()
-- end display module

-- shutdown handling module
shutdown={}
function shutdown:init(opts)
	self.reasons={}
	if not opts then
		opts={}
	end
	self.opts = opts -- note, reference not copy
	if not self.opts.lowbat_val then
		self.opts.lowbat_val = get_config_value(require'GEN/cnf_osd'.batt_volts_min)
	end
	if not self.opts.lowspace_kb_val then
		self.opts.lowspace_kb_val = 50*1024 -- 50 megs... could calc 1 raw + jpeg but not worth it
	end
end
function shutdown:check()
	if self.opts.lowbat and get_vbatt() < self.opts.lowbat_val then
		table.insert(self.reasons,'lowbat')
	end
	if self.opts.lowspace and get_free_disk_space() < self.opts.lowspace_kb_val then
		table.insert(self.reasons,'lowspace')
	end
	return #self.reasons > 0
end

function shutdown:reason()
	return table.concat(self.reasons,' ')
end

function shutdown:do_shutdown()
	post_levent_to_ui("PressPowerButton")
	-- probably not needed
	sleep(100)
	post_levent_to_ui("UnpressPowerButton")
end

function shutdown:finish()
	if #self.reasons == 0 and self.opts.finish then
		table.insert(self.reasons,'finish')
	end
	if #self.reasons > 0 then
		self:do_shutdown()
	end
end
-- end shutdown module

-- exposure module
exp = {}

function exp:init(opts)
	for i,name in pairs{
		'ev_change_max',
		'tv96_long_limit',
		'tv96_sv_thresh',
		'sv96_max',
		'sv96_target',
		'meter_size',
		'meter_step',
		'under_protect_ev_max',
		'under_protect_ev_max_thresh',
		'over_margin_ev',
		'over_thresh_pct',
		'under_margin_ev',
		'under_thresh_pct',
		'do_draw',
	} do
		if not opts[name] then
			error('exp missing opt '..name)
		end
		self[name] = opts[name]
	end

	-- meter rectangle
	self.meter_left = rawop.fb.width/2 - self.meter_size/2
	self.meter_count = self.meter_size/self.meter_step
	self.meter_top = rawop.fb.height/2 - self.meter_size/2

	-- histo limits in shot histo unites
	-- lowest value to count as over exp, as shot histo value
	-- = raw(ev(whitelevel) - margin_ev)/(2^(bpp - histo_bpp))
	self.over_histo_min = rawop.ev96_to_raw(rawop.raw_to_ev96(rawop.fb.white_level)-self.over_margin_ev)/bitshl(1,rawop.fb.bits_per_pixel - 10)
	-- highest value to count as under exp, as shot histo value 
	-- = raw(-margin_ev)/(2^(bpp - histo_bpp))
	self.under_histo_max = rawop.ev96_to_raw(-self.under_margin_ev*96)/bitshl(1,rawop.fb.bits_per_pixel - 10)

	-- used to check over/under
	-- TODO could only set if under/over pct > 0
	shot_histo_enable(1)
end

-- draw debug stuff on raw
function exp:draw()
	local t0=get_tick_count()
	local drawval_high = rawop.fb.white_level - rawop.fb.black_level
	local drawval_low = rawop.fb.black_level*2
	local drawval = drawval_high
	if self.mval > rawop.fb.white_level - rawop.fb.white_level/3 then
		drawval = drawval_low
	end
	local size = self.meter_count*self.meter_step

	-- box around meter area
	rawop.rect_rgb(self.meter_left,self.meter_top,size,size,2,drawval,drawval,drawval)
	
	-- meter level
	local bl_offset = -rawop.raw_to_ev96(rawop.fb.black_level+1)
	local m_width = ((bl_offset + self.mval96) * size)/(bl_offset + rawop.raw_to_ev96(rawop.fb.white_level))
	rawop.fill_rect_rgb(self.meter_left,
						self.meter_top-16,
						m_width,
						8,
						drawval,drawval,drawval)

	-- exposure change for next shot
	local ev_width = (self.ev_change * size/2)/self.ev_change_max
	if self.ev_change > 0 then
		rawop.fill_rect_rgb(self.meter_left+size/2,self.meter_top + size + 8,
				ev_width,8,drawval,drawval,drawval)
	elseif self.ev_change < 0 then
		ev_width = -ev_width
		rawop.fill_rect_rgb(self.meter_left+size/2 - ev_width,
							self.meter_top + size + 8,
							ev_width,
							8,drawval,drawval,drawval)
	end

	-- under exposed %
	local h = (self.under_pct*size)/100
	local r,g,b = drawval,drawval,drawval
	if self.under_pct >= self.under_thresh_pct then
		r = drawval_high
		g,b = drawval_low,drawval_low
	end
	rawop.fill_rect_rgb(self.meter_left - 16,self.meter_top + size - h,8,h,r,g,b)

	-- over exposured %
	h = (self.over_pct*size)/100
	if self.over_pct >= self.over_thresh_pct then
		r = drawval_high
		g,b = drawval_low,drawval_low
	else
		r,g,b = drawval,drawval,drawval
	end
	rawop.fill_rect_rgb(self.meter_left + size + 8,self.meter_top,8,h,r,g,b)

	log:set{draw_time=get_tick_count()-t0}
end

-- calculate desired exposure change for next shot
function exp:calc_ev_change()
	local last_ev_change
	-- handle first iteration
	if self.ev_change then
		last_ev_change = self.ev_change
	else
		last_ev_change = 0
	end

	-- basic change: difference of last exposure from metered exposure
	-- TODO might want to use initial scene metered value as target instead, or support exposure comp/shifts
	local ev_change = -self.mval96

	-- protect against over / under
	-- ramp in changes to avoid flapping, will be clamped below
	-- TODO highly dependendant on max EV step
	-- if value is exactly equal to limit specify zero change, to reduce flapping
	if self.over_pct >= self.over_thresh_pct then
		local over_ev = -self.ev_change_max*((self.over_pct - self.over_thresh_pct)*100)/(100*self.over_thresh_pct)
		logdesc('over:%d',over_ev)
		-- if histo based change (including 0) is more than meter based, use it
		-- meter might be going in the right direction anyway
		if over_ev < ev_change then
			logdesc('histo adj-')
			ev_change = over_ev
		end
	-- only worry about under if not over exposed
	elseif self.under_pct >= self.under_thresh_pct then
		local under_ev = self.ev_change_max*((self.under_pct - self.under_thresh_pct)*100)/(100*self.under_thresh_pct)
		logdesc('under:%d',under_ev)
		-- if meter is too far over target, don't try to compensate for under
		-- meter will start to bring down, may be abrupt if ev_change_max is high
		-- could set zero or small neg instead
		if self.mval96 > self.under_protect_ev_max then
			logdesc('under ignore:%d',under_ev)
		else
			-- margin before hard limit
			local margin = self.under_protect_ev_max - self.mval96
			if margin == 0 then
				under_ev = 0
			elseif under_ev >= margin then -- if req change would reach limit, clamp
				under_ev = margin - 1
			end
			-- if req change would put over threshold, ramp
			if self.mval96 + under_ev > self.under_protect_ev_max_thresh then
				local range = self.under_protect_ev_max - self.under_protect_ev_max_thresh
				under_ev = under_ev*margin/range
				logdesc('under limit:%d',under_ev)
			end
			-- if histo based change (including 0) is less than meter based, use it
			-- meter might be going in the right direction anyway
			if under_ev > ev_change then
				logdesc('histo adj+')
				ev_change = under_ev
			end
		end
	end

	-- clamp final change to ui max
	if ev_change > self.ev_change_max then
		logdesc('clamp+')
		ev_change = self.ev_change_max
	elseif ev_change < -self.ev_change_max then
		ev_change = -self.ev_change_max
		logdesc('clamp-')
	end

	-- smooth out rapid changes
	local ev_change_smooth = ev_change + last_ev_change
	ev_change_smooth = ev_change_smooth/2 + ev_change_smooth%2
	-- but don't allow overshoot in the wrong direction
	if (ev_change == 0 and ev_change_smooth ~= 0) 
		or (ev_change_smooth > 0 and ev_change < 0) 
		or (ev_change_smooth < 0 and ev_change > 0) then
		logdesc('smooth overshoot:%d',ev_change_smooth)
		ev_change = 0
	else
		ev_change = ev_change_smooth
	end

	self.ev_change = ev_change
end

-- read meter values from frame buffer
function exp:do_meter()
	local t0=get_tick_count()
	self.mval = rawop.meter(self.meter_left,self.meter_top,
								self.meter_count,self.meter_count,
								self.meter_step,self.meter_step)

	log:set{meter_time=get_tick_count()-t0}

	self.mval96 = rawop.raw_to_ev96(self.mval)

	-- shot histo values always scaled to 10 bit, assume white=1023, black=31
	-- ignore much lower than black, typically bad pixels
	self.over_pct = get_histo_range(self.over_histo_min,1023)
	self.under_pct = get_histo_range(4,self.under_histo_max)
end

function exp:simulate_do_meter()
	self.sim.frame = self.sim.frame + 1
	local delta_tv = self.sim.TV - self.sim.init_TV
	local delta_sv = self.sim.SV - self.sim.init_SV 
	local m = self.sim.init_bv + self.sim.data[self.sim.frame].delta_bv + delta_sv - delta_tv

	self.mval = rawop.ev96_to_raw(m) 
	self.mval96 = m -- extra conversion loses precision -- rawop.raw_to_ev96(self.mval)
	self.over_pct = self.sim.data[self.sim.frame].over_pct
	self.under_pct = self.sim.data[self.sim.frame].under_pct
end


-- these are functions to allow overriding for simulate
function exp:get_previous_exposure()
	-- AV is with ND if set by canon firmware
	return get_prop(props.TV), get_prop(props.SV), get_prop(props.AV)
end

function exp:simulate_get_previous_exposure()
	return self.sim.TV, self.sim.SV, self.sim.AV
end

function exp:set_exposure(tv96,sv96)
	set_tv96_direct(tv96)
	set_sv96(sv96)
end

function exp:simulate_set_exposure(tv96,sv96)
	self.sim.TV = tv96
	self.sim.SV = sv96
end

-- note index is zero based
function exp:extract_csv_col(line,index)
	local i=0
	local s=0
	while i < index and s do
		i = i + 1
		s=string.find(line,',',s+1,true)
		if not s then
			return
		end
	end
	return string.match(line,'([^,]*)',s+1)
end

function exp:sim_load_csv(filename)
	local fh=io.open(filename,'rb')
	if not fh then
		error('failed to open '..tostring(filename))
	end
	-- get column indexes from header
	local line=fh:read()
	local i=0
	local cols={}
	while true do
		local col_name = self:extract_csv_col(line,i)
		if not col_name then
			break
		end
		cols[col_name]=i
		i=i+1
	end
	local col_names={'av96','sv96','tv96','meter96','over_pct','under_pct'}

	local data={}
	local i=1
	for line in fh:lines() do
		data[i]={}
		for j,name in pairs(col_names) do
			data[i][name] = tonumber(self:extract_csv_col(line,cols[name]))
		end
		i=i+1
	end
	fh:close()
	return data
end

function exp:sim_init()
	self.sim = {
		data={},
		frame=0,
	}
	local data=self:sim_load_csv('A/rawopsim.csv')
	print('loaded',#data,'rows')
	self.sim.init_TV = data[1].tv96
	self.sim.init_SV = data[1].sv96
	self.sim.init_AV = data[1].av96
	self.sim.init_bv = data[1].meter96

	for i,v in ipairs(data) do
		self.sim.data[i] = {
			over_pct = v.over_pct,
			under_pct = v.under_pct,
			delta_bv = v.tv96 - self.sim.init_TV + self.sim.init_SV - v.sv96 + v.meter96,
		}
	end

	self.sim.TV = self.sim.init_TV
	self.sim.SV = self.sim.init_SV
	self.sim.AV = self.sim.init_AV

	-- override functions with sim versions
	for i,name in ipairs{'do_meter','get_previous_exposure','set_exposure'} do
		self[name] = self['simulate_'..name]
	end
	self.do_draw = false
end

-- meter and update exposure
function exp:run()
	self:do_meter()
	-- log meter values
	log:set{
		meter=self.mval,
		meter96=self.mval96,
		over_pct=self.over_pct,
		under_pct=self.under_pct,
	}

	if self.mval == 0 then
		log:set{meter='fail'}
		return
	end

	self:calc_ev_change()
	-- log ev change
	log:set{
		d_ev=self.ev_change,
	}

	-- desired exposure change is calculated, now set up exposure
	local tv96,sv96,av96 = self:get_previous_exposure()

	-- start with ev change on Tv
	local tv96_new = tv96 - self.ev_change

	local sv_extra = 0

	local sv96_new = self.sv96_target
	if sv96 > sv96_new then
		sv_extra = sv96 - sv96_new
	end

	-- put anything over base ISO on Tv, will add back later if tv limits hit
	tv96_new = tv96_new - sv_extra

	local tv_extra = 0

	if tv96_new < self.tv96_long_limit then
		tv_extra = tv96_new - self.tv96_long_limit
		tv96_new = self.tv96_long_limit
		logdesc('tv long limit:%d',tv_extra)
	-- over threshold, start ramping up ISO
	elseif tv96_new < self.tv96_sv_thresh then
		tv_extra = (tv96_new - self.tv96_sv_thresh)/2
		tv96_new = tv96_new - tv_extra
		logdesc('tv iso adj:%d',tv_extra)
	end

	if tv_extra < 0 then
		sv96_new = sv96_new - tv_extra
	end
	if sv96_new > self.sv96_max then
		logdesc('iso over limit:%d',sv96_new - self.sv96_max)
		sv96_new = self.sv96_max
	end

	self:set_exposure(tv96_new,sv96_new)
	
	-- exposure calculation complete, log
	local tvus=tv96_to_usec(tv96)
	local av1k=av96_to_aperture(av96)
	log:set{
		-- note exposure params logged are for the *metered* shot, not the shot just calculated!
		sv=sv96_to_iso(sv96_real_to_market(sv96)),
		sv96=sv96,
		av=string.format("%d.%03d",av1k/1000,((av1k%1000))),
		av96=av96,
		tv=string.format("%d.%06d",tvus/1000000,((tvus%1000000))),
		tv96=tv96,
	}
	if self.do_draw then
		self:draw()
	end
end

function restore()
	disp:enable(true);
	log:close()
end

-- main script initialization
if not ui_sim then
	local rec, vid = get_mode()
	if not rec then
		print("switching to rec")
		sleep(1000)
		set_record(true)
		repeat sleep(10) until get_mode()
		sleep(500)
		rec, vid = get_mode()
	end
	if vid then
		error('not in still mode')
	end
end

log:init{
	name="A/rawopint.csv",
	append=ui_log_append,
	-- column names
	cols={
		'date',
		'tick',
		'exp',
		'start',
		'shoot_ready',
		'sleep',
		'exp_start',
		'raw_ready',
		'meter_time',
		'draw_time',
		'raw_done',
		'vbatt',
		'tsensor',
		'topt',
		'free_mem',
		'lua_mem',
		'sd_space',
		'sv',
		'sv96',
		'tv',
		'tv96',
		'av',
		'av96',
		'meter',
		'meter96',
		'over_pct',
		'under_pct',
		'd_ev',
		'desc',
	},
	-- columns automatically set at write time from functions
	funcs={
		date=function()
			return os.date('%m/%d/%Y %H:%M:%S')
		end,
		tick=get_tick_count,
		exp=get_exp_count,
		vbatt=get_vbatt,
		tsensor=function()
			return get_temperature(1)
		end,
		topt=function()
			return get_temperature(0)
		end,
		free_mem=function()
			return get_meminfo().free_size
		end,
		lua_mem=function()
			return collectgarbage('count')
		end,
		sd_space=get_free_disk_space,
	},
	-- columns collected in a table, concatenated at write time
	tables={
		desc=' / ',
	},
}
logtime=log:dt_logger('start')
logdesc=log:text_logger('desc')

shutdown:init{
	finish=ui_shutdown_finish,
	lowbat=ui_shutdown_lowbat,
	lowspace=ui_shutdown_lowspace,
}

exp:init{
	meter_size=ui_meter_size,
	meter_step=ui_meter_step,

	ev_change_max=96/ui_ev_change_f,

	-- max (longest) shutter value
	tv96_long_limit=usec_to_tv96(ui_tv_max_s100*10000),
	-- shutter value to start adjusting iso
	tv96_sv_thresh=usec_to_tv96(ui_tv_sv_adj_s100*10000),

-- 1/16000th
-- TODO not implemented
--tv96_short_limit=1341

	-- max iso
	sv96_max=sv96_market_to_real(iso_to_sv96(ui_sv_max_mkt)),
	-- target iso
	sv96_target=sv96_market_to_real(iso_to_sv96(ui_sv_target_mkt)),

-- prefer low or high aperture
-- TODO not implemented
--av_target_low = (av_target.value == 0)


	-- TODO these values should be user configurable
	-- maximum meter value for under protection (1.75 stops over neutral)
	under_protect_ev_max = 168,
	-- threshold to start ramping out under protect (1.3 stops)
	under_protect_ev_max_thresh = 128,
	
    -- how close to max shot histo to count against over exp %
	over_margin_ev=96/ui_exp_over_margin_ev_f,
  
    -- under is defined in terms of EV under neutral, since there are a bunch of stops without useful DR near black level
	under_margin_ev=ui_exp_under_margin_ev, 

	over_thresh_pct=ui_exp_over_thresh_pct,
	under_thresh_pct=ui_exp_under_thresh_pct,
	do_draw=ui_do_draw,
}

function run_sim()
	exp:sim_init()
	local count,ms=set_yield(-1,50)
	local i=1
	local total = #exp.sim.data
	repeat
		print("sim",i,"/",total)
		wait_click(10)
		if is_key('menu') then
			-- prevent shutdown on finish if user abort
			shutdown.opts.finish = false
			logdesc('user exit')
			log:write()
			break
		end
		exp:run()
		log:write()
		collectgarbage('step')
		i=i+1
	until not exp.sim.data[exp.sim.frame+1]
	set_yield(count,ms)
end

function run()
	-- set initial display state
	disp:update()

	-- set the hook just before shutter release for timing
	hook_shoot.set(10000)
	-- set hook in raw for exposure
	hook_raw.set(10000)

	press('shoot_half')

	repeat sleep(10) until get_shooting()

	for i=1,ui_shots do
		print("shot",i,"/",ui_shots)
		-- poll / reset click state
		-- camera will generally take while to be ready for next shot, so extra wait here shouldn't hurt
		wait_click(10)
		if is_key('menu') then
			-- prevent shutdown on finish if user abort
			shutdown.opts.finish = false
			logdesc('user exit')
			log:write()
			break
		end
		if shutdown:check() then
			logdesc('shutdown:%s',shutdown:reason())
			log:write()
			break
		end

		press('shoot_full_only')
		local t_start=get_tick_count()
		log:set{start=t_start}
		-- wait until the hook is reached
		hook_shoot.wait_ready()
		logtime('shoot_ready')
		release('shoot_full_only')
		-- if additional wait is needed to reach the desired interval, wait
		if shot_tick then
			local et = get_tick_count() - shot_tick
			if et < interval then
				sleep(interval - et)
			end
			log:set{sleep=interval-et} -- negative == late
		end
		-- record time
		shot_tick = get_tick_count()
		logtime('exp_start') -- the moment the exposure started, because hey, why not?
		-- allow shooting to proceed
		hook_shoot.continue()

		-- TODO CHDK osd doesn't seem to update in halfshoot anyway
		if is_key('set') then
			disp:toggle(30000)
		end
		disp:update()

		-- wait for the image to be captured
		hook_raw.wait_ready()
		logtime('raw_ready')
		exp:run()
		hook_raw.continue()
		logtime('raw_done')
		log:write()
		-- encourage garbage collection at a predictable point
		-- TODO should do full collect in sleep time if avail, otherwise step
		collectgarbage('step')
	end
	release('shoot_full')
end

if ui_sim then
	run_sim()
else
	run()
end

restore()
shutdown:finish()
